"""
Core functions used in the processing of output for fmdt: https://github.com/alsoc/fmdt

Public API:
    extract_key_information
    extract_all_information
    split_video_at_meteors
"""

import fmdt.utils as utils

# Structure of tracking output table of fmdt-detect after 
# each line gets stripped using whitespace as a delimiter
__OBJECT_ID_COLUMN   = 0
__START_FRAME_COLUMN = 2
__START_X_COLUMN     = 4
__START_Y_COLUMN     = 6
__END_FRAME_COLUMN   = 8
__END_X_COLUMN       = 10
__END_Y_COLUMN       = 12
__OBJECT_TYPE_COLUMN = 14

def extract_key_information(detect_tracks_in: str) -> list[dict]:
    """Extract key information from a detect_tracks.txt file.

    "Key" information refers to the start frame, end frame, and type of 
    object detected

    Parameters
    ----------
    detect_tracks_in (str): The name of a file whose content is the output
        of fmdt_detect

    Returns
    -------
    dict_array (list[dict]): A list of dictionaries of the form
        { 
            "type": <"meteor" | "noise" | "start">
            "frame_start": <int>
            "frame_end": <int>
        }
        where each item in the list corresponds to a single object detected by 
        `fmdt-detect`
    """

    # Utility function to convert a line of interest to a dictionary
    def line_to_dict(split_line: list[str]):

        return {
            "type":         split_line[__OBJECT_TYPE_COLUMN],
            "start_frame":  int(split_line[__START_FRAME_COLUMN]),
            "end_frame":    int(split_line[__END_FRAME_COLUMN])
        }

    # Utility boolean function to extract only the important lines
    interesting_line = lambda line: (" meteor" in line) or (" star" in line) or (" noise" in line)
    
    # Processing of the actual file
    file_tracks = open(detect_tracks_in)
    file_lines  = file_tracks.readlines()
    dict_array = []

    for line in file_lines:
        if interesting_line(line):
            dict_array.append(line_to_dict(line.split()))

    return dict_array

def extract_all_information(detect_tracks_in: str) -> list[dict]:
    """Extract all tracking information from a detect_tracks.txt file.

    Parameters
    ----------
    detect_tracks_in (str): The name of a file whose content is the output
        of fmdt_detect

    Returns
    -------
    dict_array (list[dict]): A list of dictionaries of the form
        { 
            "id":           <int>,
            "start_frame":  <int>,
            "start_x":      <float>,
            "start_y":      <float>,
            "end_frame":    <int>,
            "end_x":        <float>,
            "end_y":        <float>,
            "type":         <"meteor" | "noise" | "start">,
        }
        where each item in the list corresponds to a single object detected by 
        `fmdt-detect`
    """

    # Utility function to convert a line of interest to a dictionary
    def line_to_dict(split_line: list[str]):

        return {
            "id":           int(split_line[__OBJECT_ID_COLUMN]),
            "start_frame":  int(split_line[__START_FRAME_COLUMN]),
            "start_x":      float(split_line[__START_X_COLUMN]),
            "start_y":      float(split_line[__START_Y_COLUMN]),
            "end_frame":    int(split_line[__END_FRAME_COLUMN]),
            "end_x":        float(split_line[__END_X_COLUMN]),
            "end_y":        float(split_line[__END_Y_COLUMN]),
            "type":         split_line[__OBJECT_TYPE_COLUMN]
        }

    # Utility boolean function to extract only the important lines
    interesting_line = lambda line: (" meteor" in line) or (" star" in line) or (" noise" in line)
    
    # Processing of the actual file
    file_tracks = open(detect_tracks_in)
    file_lines  = file_tracks.readlines()
    dict_array = []

    for line in file_lines:
        if interesting_line(line):
            dict_array.append(line_to_dict(line.split()))

    return dict_array

def split_video_at_meteors(video_filename: str, detect_tracks_in: str, nframes_before=3, nframes_after=3, overwrite=False, exact_split: bool = False, log: bool = False) -> None:
    """
    Split a video into small segments of length (nframes_before + nframes_after + sequence_length) frames
    for each meteor detected 

    Parameters
    ----------
    video_filename (str): Filename of video to split 
    detect_tracks_in (str): Filename of tracks recorded by a call to 
        fmdt.detect(video_filename, out_track_file=detect_tracks_in)
    nframes_before (int): Number of frames to extract before the meteor sequence begins
        (Default value of 3)
    nframes_after (int): Number of frames to extract after the meteor sequence ends
        (Default value of 3)
    overwrite (bool): Tells ffmpeg to overwrite (True) the generated output videos if they already
        exist. If false then ffmpeg will ask for manual confirmation
        (Default value of False)
    exact_split (bool): Determine whether to extract the _exact_ frames requested (True) - which requires loading
        the entire video into memory as a numpy array - or approximate the frames requested using ffmpeg's 
        seeking functionality (False). For long videos with high resolution, use False otherwise the program 
        might crash. 
        (Default value of False)
    log (bool): When True, print to console the current action being performed
    """
    utils.assert_file_exists(video_filename)
    utils.assert_file_exists(detect_tracks_in)

    # Preprocessing of information held in `detect_tracks_in`
    tracking_list = extract_key_information(detect_tracks_in)
    tracking_list = utils.retain_meteors(tracking_list)
    seqs = utils.separate_meteor_sequences(tracking_list)
    video_name, extension = utils.decompose_video_filename(video_filename) 

    # Bookkeeping for formatting the name of the output videos
    max_digits = len(str(seqs[-1][1]))
    format_str = '0' + str(max_digits)

    # Selection of splitting algorithm
    if exact_split:
        
        # Querying of video information, extraction of frames
        frames = utils.convert_video_to_ndarray(video_filename, log=log)
        frame_rate = utils.get_avg_frame_rate(video_filename)
        total_frames, _, _, _ = frames.shape
        seq_video_name = lambda seq: f'{video_name}_f{format(seq[0], format_str)}-{format(seq[1], format_str)}.{extension}'

        def exact_splitting(seq) -> None:

            # Ensure that f_start and f_end are valid
            f_start = s[0] - nframes_before if s[0] - nframes_before >= 0 else 0
            f_end   = s[1] + nframes_after  if s[1] + nframes_after  <= total_frames else total_frames
            frames_seq = frames[f_start:f_end, :, :, :]
            utils.convert_ndarray_to_video(seq_video_name(s), frames_seq, frame_rate, log=log)

        splitting_algorithm = exact_splitting
    
    else:

        seq_video_name = lambda seq: f'{video_name}_f{format(seq[0], format_str)}-{format(seq[1], format_str)}_.{extension}'

        def approx_splitting(seq) -> None:

            f_start = s[0] - nframes_before if s[0] - nframes_before >= 0 else 0
            f_end   = s[1] + nframes_after
            utils.extract_video_frames(video_filename, f_start, f_end, seq_video_name(s), overwrite=overwrite)

        splitting_algorithm = approx_splitting

    for s in seqs:
        splitting_algorithm(s)


"""
API to call fmdt executables. Assumes that fmdt-detect and other executables
are on the system path
"""

import shutil
import subprocess
import fmdt.core
import fmdt.utils
import fmdt.args

def count(trk_bb_path: str, stars: bool = False, meteors: bool = True, noise: bool = False, all: bool = False) -> int:
    """Count the number of meteors detected in a tracks_bb file"""
    if all:
        stars = True
        meteors = True
        noise = True

    tracked = fmdt.core.extract_key_information(trk_bb_path)

    if not stars:
        tracked = [t for t in tracked if t["type"] != "star"]

    if not meteors:
        tracked = [t for t in tracked if t["type"] != "meteor"]

    if not noise:
        tracked = [t for t in tracked if t["type"] != "noise"]

    # tracked_meteors = fmdt.utils.retain_meteors(tracked)
    return len(tracked)
    

def detect(
        vid_in_path: str, 
        vid_in_start: int | None = None,
        vid_in_stop: int | None = None,
        vid_in_skip: int | None = None,
        vid_in_buff: bool | None = None,
        vid_in_loop: int | None = None,
        vid_in_threads: int | None = None,
        light_min: int | None = None,
        light_max: int | None = None,
        ccl_fra_path: str | None = None,
        ccl_fra_id: bool | None = None,
        mrp_s_min: int | None = None,
        mrp_s_max: int | None = None,
        knn_k: int | None = None,
        knn_d: int | None = None,
        knn_s: int | None = None,
        trk_ext_d: int | None = None,
        trk_ext_o: int | None = None,
        trk_angle: float | None = None,
        trk_star_min: int | None = None,
        trk_meteor_min: int | None = None,
        trk_meteor_max: int | None = None,
        trk_ddev: float | None = None,
        trk_all: bool | None = None,
        trk_bb_path: str | None = None,
        trk_mag_path: str | None = None,
        log_path: str | None = None,
        out_track_file: str | None = None,
        log: bool = False
    ) -> fmdt.args.Args:

    # Wrap up all of the arguments into a dictionary
    detect_args = fmdt.args.detect_args(vid_in_path, vid_in_start, vid_in_stop, 
        vid_in_skip, vid_in_buff, vid_in_loop, vid_in_threads, light_min, light_max, 
        ccl_fra_path, ccl_fra_id, mrp_s_min, mrp_s_max, knn_k, knn_d, knn_s, trk_ext_d,
        trk_ext_o, trk_angle, trk_star_min, trk_meteor_min, trk_meteor_max, trk_ddev, 
        trk_all, trk_bb_path, trk_mag_path, log_path, out_track_file, log
    )

    # Spit out the commandline
    args = fmdt.args.handle_detect_args(**detect_args)

    if out_track_file is None:
        if log:
            print(f"Executing cmd: {' '.join(args)}")
        subprocess.run(args)
    else:
        with open(out_track_file, 'w') as outfile:
            if log:
                print(f"Executing cmd: {' '.join(args)}")
            subprocess.run(args, stdout=outfile)
    
    # And return the Args object that keeps track of the detect call

    


def visu(in_video: str, in_track_file: str, in_bb_file: str, out_visu_file: str | None = None, show_id: bool = False) -> None:

    fmdt_visu_exe = shutil.which("fmdt-visu")
    fmdt_visu_found = not fmdt_visu_exe is None
    assert fmdt_visu_found, "fmdt-visu executable not found"   

    args = [fmdt_visu_exe, "--in-video", in_video, "--in-tracks", in_track_file, "--in-bb", in_bb_file] 

    if not out_visu_file is None:
        args.extend(["--out-video", out_visu_file])

    if show_id:
        args.append("--show-id")

    subprocess.run(args)


"""Arguments Class to store shared parameters when calling a chain of .detect(args).visu().split()"""
import fmdt.api
import shutil

class Args:

    def __init__(self, tracking_list, detect_args: dict | None = None):
        self.tracking_list = tracking_list
        self.detect_args = detect_args
    
    @classmethod
    def detect(self):

        # Make sure the detecting arguments are not none
        if self.detect_args is None:
            self.detect_args = self.default_detect_args()
        
        # Convert the list of detect args to a single command

        return fmdt.api.detect(**self.detect_args)

    def visu(
            self,
            vid_in_path: str | None = None,
            vid_in_start: int | None = None,
            vid_in_stop:  int | None = None,
            vid_in_threads: int | None = None,
            trk_path: str | None = None,
            trk_bb_path: str | None = None,
            trk_id: bool | None = None,
            trk_nat_num: bool | None = None,
            trk_only_meteor: bool | None = None,
            gt_path: str | None = None,
            vid_out_path: str | None = None
        ):

        print("Not yet implemented")
    


def detect_args(
    vid_in_path: str, 
    vid_in_start: int | None = None,
    vid_in_stop: int | None = None,
    vid_in_skip: int | None = None,
    vid_in_buff: bool | None = None,
    vid_in_loop: int | None = None,
    vid_in_threads: int | None = None,
    light_min: int | None = None,
    light_max: int | None = None,
    ccl_fra_path: str | None = None,
    ccl_fra_id: bool | None = None,
    mrp_s_min: int | None = None,
    mrp_s_max: int | None = None,
    knn_k: int | None = None,
    knn_d: int | None = None,
    knn_s: int | None = None,
    trk_ext_d: int | None = None,
    trk_ext_o: int | None = None,
    trk_angle: float | None = None,
    trk_star_min: int | None = None,
    trk_meteor_min: int | None = None,
    trk_meteor_max: int | None = None,
    trk_ddev: float | None = None,
    trk_all: bool | None = None,
    trk_bb_path: str | None = None,
    trk_mag_path: str | None = None,
    log_path: str | None = None,
    out_track_file: str | None = None,
    log: bool = False
) -> dict:

    detect_args = {
        "vid_in_path": vid_in_path, 
        "vid_in_start": vid_in_start,
        "vid_in_stop": vid_in_stop,
        "vid_in_skip": vid_in_skip,
        "vid_in_buff": vid_in_buff,
        "vid_in_loop": vid_in_loop,
        "vid_in_threads": vid_in_threads,
        "light_min": light_min,
        "light_max": light_max,
        "ccl_fra_path": ccl_fra_path,
        "ccl_fra_id": ccl_fra_id,
        "mrp_s_min": mrp_s_min,
        "mrp_s_max": mrp_s_max,
        "knn_k": knn_k,
        "knn_d": knn_d,
        "knn_s": knn_s,
        "trk_ext_d": trk_ext_d,
        "trk_ext_o": trk_ext_o,
        "trk_angle": trk_angle,
        "trk_star_min": trk_star_min,
        "trk_meteor_min": trk_meteor_min,
        "trk_meteor_max": trk_meteor_max,
        "trk_ddev": trk_ddev,
        "trk_all": trk_all,
        "trk_bb_path": trk_bb_path,
        "trk_mag_path": trk_mag_path,
        "log_path": log_path,
        "out_track_file": out_track_file 
    }

    return detect_args


def default_detect_args() -> dict:       
    # Hi 
    default_detect = {
        "vid_in_path": None, 
        "vid_in_start": None,
        "vid_in_stop": None,
        "vid_in_skip": None,
        "vid_in_buff": None,
        "vid_in_loop": None,
        "vid_in_threads": None,
        "light_min": None,
        "light_max": None,
        "ccl_fra_path": None,
        "ccl_fra_id": None,
        "mrp_s_min": None,
        "mrp_s_max": None,
        "knn_k": None,
        "knn_d": None,
        "knn_s": None,
        "trk_ext_d": None,
        "trk_ext_o": None,
        "trk_angle": None,
        "trk_star_min": None,
        "trk_meteor_min": None,
        "trk_meteor_max": None,
        "trk_ddev": None,
        "trk_all": None,
        "trk_bb_path": None,
        "trk_mag_path": None,
        "log_path": None,
        "out_track_file": None
    }

    return default_detect

def handle_detect_args(

    vid_in_path: str, 
    vid_in_start: int | None = None,
    vid_in_stop: int | None = None,
    vid_in_skip: int | None = None,
    vid_in_buff: bool | None = None,
    vid_in_loop: int | None = None,
    vid_in_threads: int | None = None,
    light_min: int | None = None,
    light_max: int | None = None,
    ccl_fra_path: str | None = None,
    ccl_fra_id: bool | None = None,
    mrp_s_min: int | None = None,
    mrp_s_max: int | None = None,
    knn_k: int | None = None,
    knn_d: int | None = None,
    knn_s: int | None = None,
    trk_ext_d: int | None = None,
    trk_ext_o: int | None = None,
    trk_angle: float | None = None,
    trk_star_min: int | None = None,
    trk_meteor_min: int | None = None,
    trk_meteor_max: int | None = None,
    trk_ddev: float | None = None,
    trk_all: bool | None = None,
    trk_bb_path: str | None = None,
    trk_mag_path: str | None = None,
    log_path: str | None = None,
    out_track_file: str | None = None,
    log: bool = False

) -> list[str]:

    fmdt_detect_exe = shutil.which("fmdt-detect")
    fmdt_detect_found = not fmdt_detect_exe is None
    assert fmdt_detect_found, "fmdt-detect executable not found"
    args = [fmdt_detect_exe, "--in-video", vid_in_path]

    if not vid_in_skip is None:
        args.extend(["--vid-in-skip", vid_in_skip])

    if vid_in_buff:
        args.append("--vid-in-buff")

    if not vid_in_threads is None:
        args.extend(["--vid-in-threads", vid_in_threads])

    if not light_min is None:
        args.extend(["--ccl-hyst-lo", light_min])

    if not light_max is None:
        args.extend(["--ccl-hyst-hi", light_max])

    if not ccl_fra_path is None:
        args.extend(["--ccl_fra_path", ccl_fra_path])

    if not ccl_fra_id is None:
        args.extend(["--ccl-fra-id"])

    if not mrp_s_min is None:
        args.extend(["--mrp-s-min", mrp_s_min])

    if not mrp_s_max is None:
        args.extend(["--mrp-s-max", mrp_s_max])

    if not knn_k is None:
        args.extend(["--knn-k", knn_k])

    if not knn_d is None:
        args.extend(["--knn-d", knn_d])

    if not knn_s is None:
        args.extend(["--knn-s", knn_s])

    if not trk_ext_d is None:
        args.extend(["--trk-ext-d", trk_ext_d])

    if not trk_ext_o is None:
        args.extend(["--trk-ext-o", trk_ext_o])

    if not trk_star_min is None:
        args.extend(["--trk-star-min", trk_star_min])

    if not trk_meteor_min is None:
        args.extend(["--trk-meteor-min", trk_meteor_min])

    if not trk_meteor_max is None:
        args.extend(["--trk-meteor-max", trk_meteor_max])

    if not trk_ddev is None:
        args.extend(["--trk-ddev", trk_ddev])

    if trk_all:
        args.append("--trk-all")

    if not trk_mag_path is None:
        args.extend(["--trk-mag-path", trk_mag_path])

    if not log_path is None:
        args.extend(["--log-path", log_path])

    if not trk_angle is None:
        args.extend(["--trk-angle", trk_angle])

    if not trk_star_min is None:
        args.extend(["--trk-star-min", trk_star_min])

    if not vid_in_start is None:
        args.extend(["--vid-in-start", vid_in_start])

    if not vid_in_stop is None:
        args.extend(["--vid-in-stop", vid_in_stop])

    if not trk_bb_path is None:
        args.extend(["--trk-bb-path", trk_bb_path])

    if not vid_in_loop is None:
        args.extend(["--vid-in-loop", vid_in_loop])

    if not log_path is None:
        args.extend(["--log-path", log_path])

    return args

def detect_args_to_cmd(args: dict) -> list[str]:
    return handle_detect_args(**args)            

    
